// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "UILabelNew.hpp"
#include "game/DawnGame.hpp"

using namespace Dawn;

void _uiLabelBake(
  struct UILabelFontDef &fontDef,
  FT_Face &face,
  uint32_t fontSize
) {
  fontDef.face = &face;
  fontDef.fontSize = fontSize;

  if(FT_Set_Pixel_Sizes(face, 0, fontSize)) {
    assertUnreachable();
  }

  size_t w = 0, h = 0;
  FT_ULong c;

  for(c = NEW_LABEL_CHAR_BEGIN; c < NEW_LABEL_CHAR_END; c++) {
    // Load the character
    if(FT_Load_Char(face, c, FT_LOAD_RENDER | FT_LOAD_MONOCHROME)) {
      assertUnreachable();
    }

    // Update the width and height
    w = mathMax<size_t>(w, face->glyph->bitmap.width);
    h += face->glyph->bitmap.rows;
  }


  // Now buffer pixels to the texture
  float_t y = 0;

  // I'd love to just buffer straight to the GPU, but it seems that is a bit 
  // unstable right now.
  uint8_t *buffer = (uint8_t *)memoryAllocate(w * h * sizeof(uint8_t));
  memorySet(buffer, 0x00, sizeof(uint8_t) * w * h);

  size_t offset = 0;
  for(c = NEW_LABEL_CHAR_BEGIN; c < NEW_LABEL_CHAR_END; c++) {
    // Load the character
    if(FT_Load_Char(face, c, FT_LOAD_RENDER)) {
      assertUnreachable();
    }

    // Store the character information
    struct UILabelChar info;
    info.advanceX = face->glyph->advance.x;
    info.advanceY = face->glyph->advance.y;
    info.bitmapSize = glm::vec2(face->glyph->bitmap.width, face->glyph->bitmap.rows);
    info.bitmapPosition = glm::vec2(face->glyph->bitmap_left, -face->glyph->bitmap_top);
    info.textureY = y;
    fontDef.charStore[c] = info;

    // Buffer the pixels, oh dear GOD there has to be a more efficient way.
    for(int32_t i = 0; i < face->glyph->bitmap.rows; i++) {
      memoryCopy(
        (void *)(face->glyph->bitmap.buffer + (i * face->glyph->bitmap.width)),
        (void *)(buffer + offset),
        face->glyph->bitmap.width * sizeof(uint8_t)
      );
      offset += w * sizeof(uint8_t);
    }
    y += face->glyph->bitmap.rows;
  }

  fontDef.texture.setSize(w, h, TEXTURE_FORMAT_R);
  fontDef.texture.buffer(buffer);
  memoryFree(buffer);
}



int32_t _uiLabelQuad(
  std::string text,
  Mesh *mesh,
  struct FontShaderBufferData &bufferData,
  struct UILabelFontDef &fontDef,
  glm::vec2 &position,
  int32_t quadStart,
  int32_t partIndex
) {
  int32_t len = text.length();
  if(mesh == nullptr) return len;

  // Loop each char
  glm::vec2 wh = glm::vec2(fontDef.texture.getWidth(), fontDef.texture.getHeight());

  for(int32_t i = 0; i < len; i++) {
    int32_t j = quadStart + i;
    FT_ULong c = text[i];
    auto &charInfo = fontDef.charStore[c];

    glm::vec2 uv0 = glm::vec2(0.0f, charInfo.textureY) / wh;
    glm::vec2 uv1 = uv0 + (charInfo.bitmapSize / wh);

    QuadMesh::bufferQuadMeshWithZ(mesh,
      position + charInfo.bitmapPosition, uv0,
      position + charInfo.bitmapPosition + charInfo.bitmapSize, uv1,
      0.0f,
      j * QUAD_VERTICE_COUNT, j * QUAD_INDICE_COUNT
    );

    position.x += (float_t)(charInfo.advanceX >> 6);
    position.y += (float_t)(charInfo.advanceY >> 6);
    bufferData.fontQuadParts[j] = partIndex;
  }

  return len;
}



UILabelNew::UILabelNew(SceneItem *item) : UIComponentRenderable(item) {

}

void UILabelNew::onStart() {
  std::cout << "Hello new Label" << std::endl;
  this->shaderBuffer.init();


  if(FT_New_Face(
    getGame()->renderManager.getFontManager()->fontLibrary,
    "/usr/share/fonts/TTF/arial.ttf",
    0,
    &face
  )) {
    assertUnreachable();
  }
  _uiLabelBake(defFace, face, 32);
  

  if(FT_New_Face(
    getGame()->renderManager.getFontManager()->fontLibrary,
    "/usr/share/fonts/TTF/arialbd.ttf",
    0,
    &faceBold
  )) {
    assertUnreachable();
  }
  _uiLabelBake(defFaceBold, faceBold, 32);


  if(FT_New_Face(
    getGame()->renderManager.getFontManager()->fontLibrary,
    "/usr/share/fonts/TTF/ariali.ttf",
    0,
    &faceItalics
  )) {
    assertUnreachable();
  }
  _uiLabelBake(defFaceItalics, faceItalics, 32);

  struct FontShaderBufferData fontData;

  glm::vec2 position = glm::vec2(0, defFace.fontSize);

  testMesh.createBuffers(QUAD_VERTICE_COUNT * 96, QUAD_INDICE_COUNT * 96);
  auto x = _uiLabelQuad(
    "Hello",
    &this->testMesh,
    fontData,
    defFace,
    position,
    0,
    0
  );
  x += _uiLabelQuad(
    "World",
    &this->testMesh,
    fontData,
    defFaceItalics,
    position,
    x,
    1
  );

  fontData.fontParts[0].color = COLOR_BLUE;
  fontData.fontParts[0].texture = 0;
  fontData.fontParts[1].color = COLOR_RED;
  fontData.fontParts[1].texture = 1;
  shaderBuffer.buffer(&fontData);
}

std::vector<struct ShaderPassItem> UILabelNew::getUIRenderPasses() {
  auto canvas = this->getCanvas();
  auto shader = getGame()->renderManager.fontShader;

  struct ShaderPassItem item;
  item.shader = shader;
  item.mesh = &testMesh;
  item.matrixValues[shader->paramModel] = transform->getWorldTransform();
  item.parameterBuffers[shader->bufferUiCanvas] = &canvas->shaderBuffer;
  item.parameterBuffers[shader->bufferFont] = &this->shaderBuffer;
  item.renderFlags = RENDER_MANAGER_RENDER_FLAG_BLEND;
  item.textureSlots[0] = &defFace.texture;
  item.textureSlots[1] = &defFaceItalics.texture;
  item.textureValues[shader->paramTexture0] = 0;
  item.textureValues[shader->paramTexture1] = 1;

  return { item };
}


float_t UILabelNew::getWidth() {
  return 0;
}

float_t UILabelNew::getHeight() {
  return 0;
}

float_t UILabelNew::getContentWidth() {
  return 0;
}

float_t UILabelNew::getContentHeight() {
  return 0;
}